@@ -1,11 +1,14 @@
+import datetime
+
 import pandas as pd
 import numpy as np
 
 from sqlalchemy import *
 from sqlalchemy.exc import IntegrityError, OperationalError
 from sqlalchemy.orm import sessionmaker
+from sqlalchemy.ext.declarative import declarative_base
 
-from modules import csvExporter
+import csvExporter
 from modules.actions import transform_action_object_to_list
 from modules.global_variables import *
 from modules.last_successful import get_last_successful_actions_export
@@ -23,6 +26,7 @@ def sql_setup(logger, settings, action_or_audit):
     else:
         actions_merge = False
 
+    Base = declarative_base()
     Base.metadata.clear()
 
     if action_or_audit == 'audit':
@@ -30,78 +34,65 @@ def sql_setup(logger, settings, action_or_audit):
             table = settings[SQL_TABLE]
         else:
             table = 'iauditor_data'
-        Database = set_table(table, merge)
+        Database = set_table(table, merge, Base)
     elif action_or_audit == 'actions':
         if settings[ACTIONS_TABLE] is not None:
             table = settings[ACTIONS_TABLE]
         else:
             table = 'iauditor_actions_data'
-        ActionsDatabase = set_actions_table(table, actions_merge)
+        Database = set_actions_table(table, actions_merge, Base)
     else:
         print('No Match')
         sys.exit()
-    connection_string = '{}://{}:{}@{}:{}/{}'.format(settings[DB_TYPE],
-                                                     settings[DB_USER],
-                                                     settings[DB_PWD],
-                                                     settings[DB_SERVER],
-                                                     settings[DB_PORT],
-                                                     settings[DB_NAME])
+
+    if HEROKU_URL in settings:
+        if settings[HEROKU_URL] is not None:
+            connection_string = settings[HEROKU_URL]
+    else:
+        connection_string = '{}://{}:{}@{}:{}/{}'.format(settings[DB_TYPE],
+                                                         settings[DB_USER],
+                                                         settings[DB_PWD],
+                                                         settings[DB_SERVER],
+                                                         settings[DB_PORT],
+                                                         settings[DB_NAME])
 
     engine = create_engine(connection_string)
     meta = MetaData()
     logger.debug('Making connection to ' + str(engine))
     if action_or_audit == 'audit':
-        if not engine.dialect.has_table(engine, settings[SQL_TABLE], schema=settings[DB_SCHEMA]):
-            logger.info(settings[SQL_TABLE] + ' not Found.')
-            if settings[ALLOW_TABLE_CREATION] == 'true':
+        db_setting = settings[SQL_TABLE]
+    else:
+        db_setting = settings[ACTIONS_TABLE]
+    if not engine.dialect.has_table(engine, db_setting, schema=settings[DB_SCHEMA]):
+        logger.info(db_setting + ' not Found.')
+        if settings[ALLOW_TABLE_CREATION] == 'true':
+            Database.__table__.create(engine)
+        elif settings[ALLOW_TABLE_CREATION] == 'false':
+            logger.error('You need to create the table {} in your database before continuing. If you want the '
+                         'script '
+                         'to do it for you, set ALLOW_TABLE_CREATION to '
+                         'True in your config file'.format(db_setting))
+            sys.exit()
+        else:
+            validation = input('It doesn\'t look like a table called {} exists on your server. Would you like the '
+                               'script to try and create the table for you now? (If you\'re using '
+                               'docker, you need to set APPROVE_TABLE_CREATION to true in your config file) '
+                               '(y/n)  '.format(db_setting))
+            validation = validation.lower()
+            if validation.startswith('y'):
                 Database.__table__.create(engine)
-            elif settings[ALLOW_TABLE_CREATION] == 'false':
-                logger.error('You need to create the table {} in your database before continuing. If you want the '
-                             'script '
-                             'to do it for you, set ALLOW_TABLE_CREATION to '
-                             'True in your config file'.format(settings[SQL_TABLE]))
-                sys.exit()
             else:
-                validation = input('It doesn\'t look like a table called {} exists on your server. Would you like the '
-                                   'script to try and create the table for you now? (If you\'re using '
-                                   'docker, you need to set APPROVE_TABLE_CREATION to true in your config file) '
-                                   '(y/n)  '.format(settings[SQL_TABLE]))
-                validation = validation.lower()
-                if validation.startswith('y'):
-                    Database.__table__.create(engine)
-                else:
-                    logger.info('Stopping the script. Please either re-run the script or create your table manually.')
-                    sys.exit()
-        setup = 'complete'
-        logger.info('Successfully setup Database and connection')
-    else:
-        if not engine.dialect.has_table(engine, settings[ACTIONS_TABLE], schema=settings[DB_SCHEMA]):
-            logger.info(settings[ACTIONS_TABLE] + ' not Found.')
-            if settings[ALLOW_TABLE_CREATION] == 'true':
-                ActionsDatabase.__table__.create(engine)
-            elif settings[ALLOW_TABLE_CREATION] == 'false':
-                logger.error('You need to create the table {} in your database before continuing. If you want the '
-                             'script to do it for you, set ALLOW_TABLE_CREATION to True in your '
-                             'config file'.format(settings[SQL_TABLE]))
+                logger.info('Stopping the script. Please either re-run the script or create your table manually.')
                 sys.exit()
-            else:
-                validation = input('It doesn\'t look like a table called {} exists on your server. Would you like the '
-                                   'script to try and create the table for you now? (If you\'re using '
-                                   'docker, you need to set APPROVE_TABLE_CREATION to true in your config file) '
-                                   '(y/n)  '.format(settings[ACTIONS_TABLE]))
-                validation = validation.lower()
-                if validation.startswith('y'):
-                    ActionsDatabase.__table__.create(engine)
-                else:
-                    logger.info('Stopping the script. Please either re-run the script or create your table manually.')
-                    sys.exit()
-        setup = 'complete'
-        logger.info('Successfully setup Database and connection')
+    setup = 'complete'
+    logger.info('Successfully setup Database and connection')
 
-    if action_or_audit == 'audit':
-        return setup, engine, connection_string, meta, Database
-    else:
-        return setup, engine, connection_string, meta, ActionsDatabase
+    Session = sessionmaker(bind=engine)
+    session = Session()
+
+    return setup, session, Database
+
+    #return setup, engine, connection_string, meta, Database
 
 
 def export_audit_sql(logger, settings, audit_json, get_started):
@@ -112,35 +103,48 @@ def export_audit_sql(logger, settings, audit_json, get_started):
     :param audit_json:  Audit JSON
     :get_started:       Tuple containing settings
     """
-    engine = get_started[1]
-    database = get_started[4]
+    database = get_started[2]
+    session = get_started[1]
 
     csv_exporter = csvExporter.CsvExporter(audit_json, settings[EXPORT_INACTIVE_ITEMS_TO_CSV])
     df = csv_exporter.audit_table
     df = pd.DataFrame.from_records(df, columns=SQL_HEADER_ROW)
     df['DatePK'] = pd.to_datetime(df['DateModified']).values.astype(np.int64) // 10 ** 6
-    if settings[DB_TYPE].startswith(('mysql', 'postgres')):
-        df.replace({'DateCompleted': ''}, '1900-01-01 00:00:00', inplace=True)
-        df.replace({'ConductedOn': ''}, '1900-01-01 00:00:00', inplace=True)
+    if settings[DB_TYPE].startswith('postgres'):
+        df.replace({'DateCompleted': ''}, np.datetime64(None), inplace=True)
+        df.replace({'ConductedOn': ''}, np.datetime64(None), inplace=True)
+        empty_value = np.nan
+        empty_score = empty_value
+    elif settings[DB_TYPE].startswith('mysql'):
+        df.replace({'ItemScore': '', 'ItemMaxScore': '', 'ItemScorePercentage': ''}, 0.0, inplace=True)
+        empty_value = '1970-01-01T00:00:01'
+        df.replace({'DateCompleted': ''}, empty_value, inplace=True)
+        df.replace({'ConductedOn': ''}, empty_value, inplace=True)
         df['DateStarted'] = pd.to_datetime(df['DateStarted'])
         df['DateCompleted'] = pd.to_datetime(df['DateCompleted'])
         df['DateModified'] = pd.to_datetime(df['DateModified'])
-        df['ConductedOn'] = pd.to_datetime(df['ConductedOn'])
-    df.replace({'ItemScore': '', 'ItemMaxScore': '', 'ItemScorePercentage': ''}, np.nan, inplace=True)
-    df.fillna(0, inplace=True)
+        df['ConductedOn'] = pd.to_datetime(df['ConductedOn'], format='%Y-%m-%d %H:%M:%S', utc=False)
+        df['ConductedOn'] = df['ConductedOn'].dt.tz_localize(None)
+        empty_value = None
+        empty_score = 0.0
+    else:
+        empty_value = None
+        empty_score = empty_value
+
+    df.replace({'ItemScore': '', 'ItemMaxScore': '', 'ItemScorePercentage': ''}, empty_score, inplace=True)
+    df.replace(r'^\s*$', empty_value, regex=True, inplace=True)
     df['SortingIndex'] = range(1, len(df) + 1)
     df_dict = df.to_dict(orient='records')
-    Session = sessionmaker(bind=engine)
-    session = Session()
-
     try:
         session.bulk_insert_mappings(database, df_dict)
     except KeyboardInterrupt:
         logger.warning('Interrupted by user, exiting.')
         session.rollback()
+        session.close()
         sys.exit(0)
     except OperationalError as ex:
         session.rollback()
+        session.close()
         logger.warning('Something went wrong. Here are the details: {}'.format(ex))
     except IntegrityError as ex:
         # If the bulk insert fails, we do a slower merge
@@ -153,6 +157,23 @@ def export_audit_sql(logger, settings, audit_json, get_started):
     session.commit()
 
 
+def end_session(session):
+    session.close()
+
+
+def query_max_last_modified(session, database):
+    qry = session.query(func.max(database.DateModified).label("max"))
+    res = qry.one()
+    if not res.max:
+        return '2000-01-01T00:00:00.000Z'
+    else:
+        new_last_successful = res.max + datetime.timedelta(0, 10)
+        new_last_successful = str(new_last_successful)
+        # new_last_successful = str(res.max)
+        print(new_last_successful)
+        return new_last_successful
+
+
 def save_exported_actions_to_db(logger, actions_array, settings, get_started):
     """
     Write Actions to 'iauditor_actions.csv' on disk at specified location
